/**********************************************************************
 * 
 *  Toby Opferman
 *
 *  Driver Example
 *
 *  This example is for educational purposes only.  I license this source
 *  out for use in learning how to write a device driver.
 *
 *     Driver Entry Point
 **********************************************************************/
/**
 * Eightball Driver.
 * Written by David Lariviere
 * Fall 2006 - Hardware/Software Formal Verification
 * Term Project
 * -------------
 * Driver for an imaginary device (an eightball).
 * User programs open the device, write questions to it, and then read the response. 
 * -------------
 * Code is based off of example driver described in tutorial @ http://www.codeproject.com/system/driverdev.asp
 */
#define _X86_
#include <wdm.h>
#include "example.h"

DRIVER_UNLOAD DriverUnload; //prefast told me to add this
NTSTATUS DriverEntry(PDRIVER_OBJECT pDriverObject, PUNICODE_STRING pRegistryPath);

/* 
 * These compiler directives tell the Operating System how to load the
 * driver into memory. The "INIT" section is discardable as you only
 * need the driver entry upon initialization, then it can be discarded.
 *
 */
#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(PAGE, DriverUnload)

/*
 * Required call, instructs driver to unload (perform clean-up presumably, before doing so).
 */
void DriverUnload(PDRIVER_OBJECT pDriverObject)
{
	UNICODE_STRING usDosDeviceName;

	DbgPrint("Eightball: DriverUnload Called \r\n");

	RtlInitUnicodeString(&usDosDeviceName, L"\\DosDevices\\Eightball");
	IoDeleteSymbolicLink(&usDosDeviceName);

	IoDeleteDevice(pDriverObject->DeviceObject);
	DbgPrint("Eightball: unloading completed\n");
}

/*
 * NTSTATUS DriverEntry: first called when driver is loaded
 * Comments below for arguments have been pasted from cited tutorial article for in-code reference
 * -------
 * PDRIVER_OBJECT pDriverObject: "The DRIVER_OBJECT is a data structure used to represent this driver.
		 The DriverEntry routine will use it to populate it with other entry points to the driver for handling specific I/O requests. 
		 This object also has a pointer to a DEVICE_OBJECT which is a data structure which represents a particular device. 
		 A single driver may actually advertise itself as handling multiple devices, and as such, the DRIVER_OBJECT maintains a linked list pointer 
			to all the devices this particular driver services request for."
 * PUNICODE_STRING pRegistryPath: The Registry Path is a string which points to the location in the registry where the information for the driver was stored. The driver can use this location to store driver specific information.

 */
NTSTATUS DriverEntry(PDRIVER_OBJECT pDriverObject, PUNICODE_STRING pRegistryPath)
{
	/***ADDED FOR STATICDV BUG TRIPPING****/
	PVOID *badPointer = NULL;
    KIRQL oldIrql;
	//***END*/

	NTSTATUS NtStatus = STATUS_SUCCESS;
    UINT uiIndex = 0;
    PDEVICE_OBJECT pDeviceObject = NULL;
    UNICODE_STRING usDriverName, usDosDeviceName;
	char randString[] = "evil non-safe string";
	char destStringBuf[128]; 
	strcpy(randString, destStringBuf); 

	 /*
       This defect is injected for the "CancelSpinLock" rule.
    */
    IoAcquireCancelSpinLock(&oldIrql);
	
	/*
      This defect is injected for the "NullExFreePool" rule.
    */
    ExFreePool(badPointer);

	DbgPrint("Eightball: driver entry called\n");
	
    RtlInitUnicodeString(&usDriverName, L"\\Device\\Eightball");
    RtlInitUnicodeString(&usDosDeviceName, L"\\DosDevices\\Eightball"); 

    NtStatus = IoCreateDevice(pDriverObject, 0, &usDriverName, FILE_DEVICE_UNKNOWN, FILE_DEVICE_SECURE_OPEN, FALSE, &pDeviceObject);

    if(NtStatus == STATUS_SUCCESS)
    {
        /*
         * The "MajorFunction" is a list of function pointers for entry points into the driver.
         * You can set them all to point to 1 function, then have a switch statement for all
         * IRP_MJ_*** functions or you can set specific function pointers for each entry
         * into the driver.
         *
         */
        for(uiIndex = 0; uiIndex < IRP_MJ_MAXIMUM_FUNCTION; uiIndex++)
             pDriverObject->MajorFunction[uiIndex] = Eightball_UnSupportedFunction;
    
        pDriverObject->MajorFunction[IRP_MJ_CLOSE]             = Eightball_Close;
        pDriverObject->MajorFunction[IRP_MJ_CREATE]            = Eightball_Create;
        pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]    = Eightball_IoControl;
        pDriverObject->MajorFunction[IRP_MJ_READ]              = Eightball_ReadDirectIO;
        pDriverObject->MajorFunction[IRP_MJ_WRITE]             = USE_WRITE_FUNCTION;
		//USE_READ_FUNCTION;
        /* 
         * Required to unload the driver dynamically.  If this function is missing
         * the driver cannot be dynamically unloaded.
         */
        pDriverObject->DriverUnload =  DriverUnload; 

        /* 
         * Setting the flags on the device driver object determine what type of I/O
         * you wish to use.
         *   
         *  Direct I/O - MdlAddress describes the Virtual Address list.  This is then
         *                 used to lock the pages in memory.
         *
         *                 PROS: Fast, Pages are not copied.
         *                 CONS: Uses resources, needs to lock pages into memory.
         *
         *  Buffered I/o - SystemBuffer is then used by the driver to access the data.  The I/O
         *                   manager will copy the data given by the user mode driver into this buffer
         *                   on behalf of the driver.
         *
         *                   CONS: Slower operation (Use on smaller data sets)
         *                         Uses resources, allocates Non-paged memory
         *                         Large allocations may not work since it would
         *                         require allocating large sequential non-paged memory.
         *                   PROS: Easier to use, driver simply accesses the buffer
         *                         Usermode buffer is not locked in memory
         *
         *
         *  Neither Buffered or Direct - This is when you simply read the buffer directly using the user-mode address.
         *                     Simply omit DO_DIRECT_IO and DO_BUFFERED_IO to perform this action.
         *
         *                     PROS: No copying or locking pages occurs.
         *
         *                     CONS: You *MUST* be in the context of the user-mode thread that made the request.
         *                           being in another process space you the page tables would not point to
         *                           the same location.
         *                           You have to perform some checking and probeing in order to verify
         *                           when you can read/write from the pages.
         *                           You cannot access a user mode address unless it's locked into memory
         *                           at >= DPC level.
         *                           The usermode process could also change the access rights of the
         *                           buffer while the driver is trying to read it!
         *
         *
         *  If your driver services lower level drivers you will need to set this field to the same type of
         *  I/O.
         *
         *  The flags for Read/Write is:
         *      DO_BUFFERED_IO, DO_DIRECT_IO, Specify neither flag for "Neither".
         *
         *  The flags (defined in the IOCTL itself) for Control I/O is:
         *     METHOD_NEITHER, METHOD_BUFFERED, METHOD_IN_DIRECT or METHOD_OUT_DIRECT 
         *
         *  From MSDN:
         *    For IRP_MJ_READ and IRP_MJ_WRITE requests, drivers specify the I/O method by using flags in each 
         *    DEVICE_OBJECT structure. For more information, see Initializing a Device Object.
         *
         *    For IRP_MJ_DEVICE_CONTROL and IRP_MJ_INTERNAL_DEVICE_CONTROL requests, the I/O method is determined 
         *    by the TransferType value that is contained in each IOCTL value. For more information, see Defining
         *    I/O Control Codes.
         */
         
        pDeviceObject->Flags |= IO_TYPE;
    
        /*
         * We are not required to clear this flag in the DriverEntry as the I/O Manager will
         * clear it for us, but we will anyway.  Creating a device in any other location we
         * would need to clear it.
         */
    
        pDeviceObject->Flags &= (~DO_DEVICE_INITIALIZING);
    
        /*
         * Create a Symbolic Link to the device. Example -> \Device\Example
         */
    
        IoCreateSymbolicLink(&usDosDeviceName, &usDriverName);

		DbgPrint("Eightball: driver entry sucessfully finished\n");
    }


    return NtStatus;
}