/**********************************************************************
 * 
 *  Toby Opferman
 *
 *  Driver Example
 *
 *  This example is for educational purposes only.  I license this source
 *  out for use in learning how to write a device driver.
 *
 *     Driver Functionality
 **********************************************************************/
/**
 * Eightball Driver.
 * Modified by David Lariviere
 * Fall 2006 - Hardware/Software Formal Verification
 * Term Project
 * -------------
 * Driver for an imaginary device (an eightball).
 * User programs open the device, write questions to it, and then read the response. 
 * -------------
 * Code is derived directly from example driver described in tutorial @ http://www.codeproject.com/system/driverdev.asp
 */


#define _X86_ 

#include <wdm.h>
#include "example.h"


/**********************************************************************
 * Internal Functions
 **********************************************************************/
BOOLEAN Eightball_IsStringTerminated(PCHAR pString, UINT uiLength);


#pragma alloc_text(PAGE, Eightball_Create) 
#pragma alloc_text(PAGE, Eightball_Close) 
#pragma alloc_text(PAGE, Eightball_IoControl) 
#pragma alloc_text(PAGE, Eightball_ReadDirectIO)
#pragma alloc_text(PAGE, Eightball_WriteDirectIO)
#pragma alloc_text(PAGE, Eightball_WriteBufferedIO)
#pragma alloc_text(PAGE, Eightball_WriteNeither)
#pragma alloc_text(PAGE, Eightball_UnSupportedFunction)
#pragma alloc_text(PAGE, Eightball_IsStringTerminated)




/**********************************************************************
 * 
 *  Eightball_Create
 *
 *    This is called when an instance of this driver is created (CreateFile)
 *
 **********************************************************************/
NTSTATUS Eightball_Create(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_SUCCESS;
    PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
	DbgPrint("Eightball_Create Called \r\n");

    return NtStatus;
}

/**********************************************************************
 * 
 *  Eightball_Close
 *
 *    This is called when an instance of this driver is closed (CloseHandle)
 *
 **********************************************************************/
NTSTATUS Eightball_Close(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_SUCCESS;
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
    DbgPrint("Eightball_Close Called \r\n");

    return NtStatus;
}



/**********************************************************************
 * 
 *  Eightball_IoControl
 *
 *    This is called when an IOCTL is issued on the device handle (DeviceIoControl)
 *
 **********************************************************************/
NTSTATUS Eightball_IoControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_SUCCESS;
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
    DbgPrint("Eightball_IoControl Called \r\n");

    return NtStatus;
}

int askedQuestion = 0; 

/**********************************************************************
 * 
 *  Eightball_Read
 *
 *    This is called when a read is issued on the device handle (ReadFile/ReadFileEx)
 *
 **********************************************************************/
NTSTATUS Eightball_ReadDirectIO(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    /*
	NTSTATUS NtStatus = STATUS_SUCCESS;
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
    DbgPrint("Eightball_Read Called \r\n");
	*/
	PCHAR pEightballResponses[] = { "You didn't ask Eightball a question yet!", "It is certain", "Signs point to yes", "Outlook unclear", "Don't count on it", "Ask again later" }; 
	NTSTATUS NtStatus = STATUS_BUFFER_TOO_SMALL;
    PIO_STACK_LOCATION pIoStackIrp = NULL;
    PCHAR pReturnData = "Eightball_ReadDirectIO - Hello from the Kernel!";
    UINT dwDataSize = sizeof("Eightball_ReadDirectIO - Hello from the Kernel!");
    UINT dwDataRead = 0;
    PCHAR pReadDataBuffer;
	
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
   
    DbgPrint("Eightball_ReadDirectIO Called \r\n");
    
    /*
     * Each time the IRP is passed down the driver stack a
     * new stack location is added
     * specifying certain parameters for the IRP to the 
     * driver.
     */
    pIoStackIrp = IoGetCurrentIrpStackLocation(Irp);
	DbgPrint("Eightball_Read: passed IoGetCurrentIrpStackLocation \r\n");
    if(pIoStackIrp && Irp->MdlAddress)
    {
        pReadDataBuffer = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, 
                                   NormalPagePriority);
		DbgPrint("Eightball_Read: passed MmGetSystemAddressForMdlSafe \r\n");
        if(pReadDataBuffer && 
            pIoStackIrp->Parameters.Read.Length >= dwDataSize)
        {                             
            /*
             * We use "RtlCopyMemory" in the kernel instead
             * of memcpy.
             * RtlCopyMemory *IS* memcpy, however it's best
             * to use the
             * wrapper in case this changes in the future.
             */
			/*
            RtlCopyMemory(pReadDataBuffer, pReturnData, 
                               dwDataSize);
		   */
			/*
			if (!askedQuestion) { 
				RtlCopyMemory(pReadDataBuffer, pEightballResponses[0], strlen(pEightballResponses[0]));
			}
			else { 
				RtlCopyMemory(pReadDataBuffer, pEightballResponses[1], strlen(pEightballResponses[1]));
				askedQuestion = 0; 
			}
			*/ 
			DbgPrint("Eightball_Read: passed RtlCopyMemory \r\n");
       
            dwDataRead = dwDataSize;
            NtStatus = STATUS_SUCCESS;
        }
    }
	DbgPrint("Eightball_Read: passed pIoStackIrp && Irp->MdlAddress \r\n");
    

    Irp->IoStatus.Status = NtStatus;		//return code
    Irp->IoStatus.Information = dwDataRead;	//return value

    IoCompleteRequest(Irp, IO_NO_INCREMENT);//IO_NO_INCREMEMENT instructs kernel not to boost priority of the thread within the scheduler
	DbgPrint("Eightball_Read: passed IoCompleteRequest \r\n");
    
    return NtStatus;
}



/**********************************************************************
 * 
 *  Eightball_WriteDirectIO
 *
 *    This is called when a write is issued on the device handle (WriteFile/WriteFileEx)
 *
 *    This version uses Direct I/O
 *
 **********************************************************************/
NTSTATUS Eightball_WriteDirectIO(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_SUCCESS;
    PIO_STACK_LOCATION pIoStackIrp = NULL;
    PCHAR pWriteDataBuffer;
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
    DbgPrint("Eightball_WriteDirectIO Called \r\n");
    
    /*
     * Each time the IRP is passed down the driver stack a new stack location is added
     * specifying certain parameters for the IRP to the driver.
     */
    pIoStackIrp = IoGetCurrentIrpStackLocation(Irp);
    
    if(pIoStackIrp)
    {
        pWriteDataBuffer = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);
    
        if(pWriteDataBuffer)
        {                             
            /*
             * We need to verify that the string is NULL terminated. Bad things can happen
             * if we access memory not valid while in the Kernel.
             */
           if(Eightball_IsStringTerminated(pWriteDataBuffer, pIoStackIrp->Parameters.Write.Length))
           {
                DbgPrint(pWriteDataBuffer);
				askedQuestion = 1; 
           }
        }
    }

    return NtStatus;
}

/**********************************************************************
 * 
 *  Eightball_WriteBufferedIO
 *
 *    This is called when a write is issued on the device handle (WriteFile/WriteFileEx)
 *
 *    This version uses Buffered I/O
 *
 **********************************************************************/
NTSTATUS Eightball_WriteBufferedIO(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_SUCCESS;
    PIO_STACK_LOCATION pIoStackIrp = NULL;
    PCHAR pWriteDataBuffer;
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
    DbgPrint("Eightball_WriteBufferedIO Called \r\n");
    
    /*
     * Each time the IRP is passed down the driver stack a new stack location is added
     * specifying certain parameters for the IRP to the driver.
     */
    pIoStackIrp = IoGetCurrentIrpStackLocation(Irp);
    
    if(pIoStackIrp)
    {
        pWriteDataBuffer = (PCHAR)Irp->AssociatedIrp.SystemBuffer;
    
        if(pWriteDataBuffer)
        {                             
            /*
             * We need to verify that the string is NULL terminated. Bad things can happen
             * if we access memory not valid while in the Kernel.
             */
           if(Eightball_IsStringTerminated(pWriteDataBuffer, pIoStackIrp->Parameters.Write.Length))
           {
                DbgPrint(pWriteDataBuffer);
           }
        }
    }

    return NtStatus;
}

/**********************************************************************
 * 
 *  Eightball_WriteNeither
 *
 *    This is called when a write is issued on the device handle (WriteFile/WriteFileEx)
 *
 *    This version uses Neither buffered or direct I/O.  User mode memory is
 *    read directly.
 *
 **********************************************************************/
NTSTATUS Eightball_WriteNeither(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_SUCCESS;
    PIO_STACK_LOCATION pIoStackIrp = NULL;
    PCHAR pWriteDataBuffer;
	
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
    DbgPrint("Eightball_WriteNeither Called \r\n");
    
    /*
     * Each time the IRP is passed down the driver stack a new stack location is added
     * specifying certain parameters for the IRP to the driver.
     */
    pIoStackIrp = IoGetCurrentIrpStackLocation(Irp);
    
    if(pIoStackIrp)
    {
		/*
		* We need this in an exception handler or else we could trap.
		* Note: Added nested try/except since PREfast recommended we only process the expected possible exceptions, which in this case are just the two that ProbeForRead can throw
		*/
		__try { 
			__try {
				/*
				* WDK Docs: The ProbeForRead routine checks that a user-mode buffer actually resides in the user portion of the address space, and is correctly aligned. 
				* Exception throwing Detials:
				* If the specified range of memory is not within the user-mode address range, ProbeForRead raises the STATUS_ACCESS_VIOLATION exception. 
				* If the beginning of the address range is not aligned on the byte boundary that is specified by Alignment, ProbeForRead raises the STATUS_DATATYPE_MISALIGNMENT exception. 
				*/
				ProbeForRead(Irp->UserBuffer, pIoStackIrp->Parameters.Write.Length, TYPE_ALIGNMENT(char));
				pWriteDataBuffer = Irp->UserBuffer;

				if(pWriteDataBuffer)
				{                             
					/*
					* We need to verify that the string is NULL terminated. Bad things can happen
					* if we access memory not valid while in the Kernel.
					*/
					if(Eightball_IsStringTerminated(pWriteDataBuffer, pIoStackIrp->Parameters.Write.Length))
					{
						DbgPrint(pWriteDataBuffer);
					}
				}
			//exception handling syntax taken from Platform SDK: "Using Structured Exception Handling"
			} __except(GetExceptionCode()== STATUS_ACCESS_VIOLATION ?  EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH) { //changed from EXCEPTION_EXECUTE_HANDLER per PREfast
				NtStatus = GetExceptionCode();     
			}
		}
		__except(GetExceptionCode()== EXCEPTION_EXECUTE_HANDLER ?  EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH) {
			NtStatus = GetExceptionCode(); 
		} 


    }

    return NtStatus;
}
                       

/**********************************************************************
 * 
 *  Eightball_UnSupportedFunction
 *
 *    This is called when a major function is issued that isn't supported.
 *
 **********************************************************************/
NTSTATUS Eightball_UnSupportedFunction(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_NOT_SUPPORTED;
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory
	DbgPrint("Eightball_UnSupportedFunction Called \r\n");

    return NtStatus;
}


/**********************************************************************
 * 
 *  Eightball_IsStringTerminated
 *
 *    Simple function to determine a string is NULL terminated.
 *
 **** We could validate also the characters in the string are printable! ***
 *
 **********************************************************************/
BOOLEAN Eightball_IsStringTerminated(PCHAR pString, UINT uiLength)
{
    BOOLEAN bStringIsTerminated = FALSE;
    UINT uiIndex = 0;
	PAGED_CODE(); //added per PREFAST, represents an ASSERT in checked-mode that the following code may access paged memory

    while(uiIndex < uiLength && bStringIsTerminated == FALSE)
    {
        if(pString[uiIndex] == '\0')
        {
            bStringIsTerminated = TRUE;
        }
        else
        {
           uiIndex++;
        }
    }

    return bStringIsTerminated;
}


